//Based on circuit and sketches from Wayne at http://warmcat.uk/

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <SPI.h>

#include <LittleFS.h>
const char* fsName = "LittleFS";
FS* fileSystem = &LittleFS;
LittleFSConfig fileSystemConfig = LittleFSConfig();

#define DBG_OUTPUT_PORT Serial
#ifndef STASSID
#define STASSID "xxxxx"
#define STAPSK  "xxxxx"
#endif

const char* ssid = STASSID;
const char* password = STAPSK;
const char* host = "LED_Matrix";

ESP8266WebServer server(8080);

static bool fsOK;
String unsupportedFiles = String();
String showing_message;
int message_speed = 400;
int message_position = 0;

static const char TEXT_PLAIN[] PROGMEM = "text/plain";
static const char FS_INIT_ERROR[] PROGMEM = "FS INIT ERROR";
static const char FILE_NOT_FOUND[] PROGMEM = "FileNotFound";
static const char ONLINE_AT[] PROGMEM = "ONLINE NA:";
static const char SERIAL_AT[] PROGMEM = "RYCHLOST PORTU: 115200 BAUDU";

//Display setting
int latchPin = D7;          //Pin connected to ST_CP of 74HC595
int clockPin = D6;          //Pin connected to SH_CP of 74HC595
int dataPin = D8;           //Pin connected to DS of 74HC595
int digitSel0Pin = D3;      // Pin connected to "A0 digital select" on display
int digitSel1Pin = D0;      // Pin connected to "A1 digital select" on display
int displayWritePin = D5;   // Pin connected to "WR Write" on display
int displaysCS1pins[] = {D2, D1};
byte count_digits_per_display = 4;
byte count_displays = sizeof(displaysCS1pins) / sizeof(displaysCS1pins[0]);

unsigned long aktualniMillis; //aktualni cas
unsigned long predchoziMillis; //cas posledni akce

void replyNotFound(String msg) {
  server.send(404, FPSTR(TEXT_PLAIN), msg);
}

void replyOKWithMsg(String msg) {
  server.send(200, FPSTR(TEXT_PLAIN), msg);
}

void replyServerError(String msg) {
  DBG_OUTPUT_PORT.println(msg);
  server.send(500, FPSTR(TEXT_PLAIN), msg + "\r\n");
}

void handleSetStatus() {
  if (server.arg("new_text")) {
    showing_message = "";
    for (byte i = 0; i < count_digits_per_display * count_displays; i++) {
      showing_message += " ";
    }
    showing_message += server.arg("new_text");
    showing_message += "  ";
    message_position = 0;
  }
  if (server.arg("new_speed")) {
    message_speed = server.arg("new_speed").toInt();
  }
  replyOKWithMsg(showing_message);
}

bool handleFileRead(String path) {
  if (!fsOK) {
    replyServerError(FPSTR(FS_INIT_ERROR));
    return true;
  }

  if (path.endsWith("/")) {
    path += "index.html";
  }

  String contentType;
  if (fileSystem->exists(path)) {
    File file = fileSystem->open(path, "r");
    if (server.streamFile(file, contentType) != file.size()) {
      DBG_OUTPUT_PORT.println("Sent less data than expected!");
    }
    file.close();
    return true;
  }
  return false;
}

void handleNotFound() {
  if (!fsOK) {
    return replyServerError(FPSTR(FS_INIT_ERROR));
  }

  String uri = ESP8266WebServer::urlDecode(server.uri()); // required to read paths with blanks

  if (handleFileRead(uri)) {
    return;
  }

  String message;
  message.reserve(100);
  message = F("Error: File not found\n\nURI: ");
  message += uri;
  DBG_OUTPUT_PORT.print(message);

  return replyNotFound(message);
}

void setup(void) {

  DBG_OUTPUT_PORT.begin(115200);
  DBG_OUTPUT_PORT.setDebugOutput(true);
  DBG_OUTPUT_PORT.print('\n');

  fileSystemConfig.setAutoFormat(false);
  fileSystem->setConfig(fileSystemConfig);
  fsOK = fileSystem->begin();

  DBG_OUTPUT_PORT.println(fsOK ? F("Filesystem initialized.") : F("Filesystem init failed!"));
  DBG_OUTPUT_PORT.printf("Connecting to %s\n", ssid);

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  // Wait for connection
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    DBG_OUTPUT_PORT.print(".");
  }
  DBG_OUTPUT_PORT.println("");
  DBG_OUTPUT_PORT.print(F("Connected! IP address: "));
  DBG_OUTPUT_PORT.println(WiFi.localIP());

  if (MDNS.begin(host)) {
    MDNS.addService("http", "tcp", 80);
  }

  server.on("/setmessage", HTTP_GET, handleSetStatus);
  server.onNotFound(handleNotFound);

  // Start server
  server.begin();
  DBG_OUTPUT_PORT.println("HTTP server started");

  // Start display
  pinMode(latchPin, OUTPUT);
  pinMode(clockPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(digitSel0Pin, OUTPUT);
  pinMode(digitSel1Pin, OUTPUT);
  pinMode(displayWritePin, OUTPUT);

  for (byte j = 0; j < count_displays; j++) {
    pinMode(displaysCS1pins[j], OUTPUT);
  }

  showing_message = "        ";
  showing_message += FPSTR(ONLINE_AT);
  showing_message += "  ";
  showing_message += WiFi.localIP().toString();
  showing_message += ":8080";
  showing_message += "  ";
  showing_message += FPSTR(SERIAL_AT);
  showing_message += "  ";
  message_position = 0;
}

void displayChar(char myChar, int myPos)
{
  switch (myPos)
  {
    case 3:
      digitalWrite(digitSel0Pin, LOW);
      digitalWrite(digitSel1Pin, LOW);
      break;
    case 2:
      digitalWrite(digitSel0Pin, HIGH);
      digitalWrite(digitSel1Pin, LOW);
      break;
    case 1:
      digitalWrite(digitSel0Pin, LOW);
      digitalWrite(digitSel1Pin, HIGH);
      break;
    case 0:
      digitalWrite(digitSel0Pin, HIGH);
      digitalWrite(digitSel1Pin, HIGH);
      break;
  }
  // set shift register
  digitalWrite(latchPin, LOW);
  shiftOut(dataPin, clockPin, MSBFIRST, myChar);
  digitalWrite(displayWritePin, LOW);
  digitalWrite(latchPin, HIGH);
  digitalWrite(displayWritePin, HIGH);
}

void loop(void) {
  aktualniMillis = millis();
  if (aktualniMillis - predchoziMillis > message_speed) {
    predchoziMillis = aktualniMillis;
    if (message_position > showing_message.length()) {
      message_position = 0;
    }
    for (byte i = 0; i < count_displays ; i++) {
      digitalWrite(displaysCS1pins[i], LOW);
      for (int k = 0; k < count_digits_per_display; k++) {
        displayChar(showing_message.charAt((count_digits_per_display * i) + message_position + k) , k);
      }
      digitalWrite(displaysCS1pins[i], HIGH);
    }
    message_position += 1;
  }

  server.handleClient();
  MDNS.update();
}
